/* ====================================================================
 * Copyright (c) 2000 Dean Gaudet.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Dean Gaudet"
 *
 * THIS SOFTWARE IS PROVIDED BY DEAN GAUDET ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DEAN GAUDET OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */
#ifndef SPLIM_VFD_H
#define SPLIM_VFD_H 1

/* vfd is a "virtual file descriptor"
 * it is intended to simplify the implementation of transport layers
 * in a non-blocking progamming model by providing an abstraction of
 * the relevant portions of the unix socket API -- specifically,
 * read, write, and select.
 *
 * all vfds are non-blocking.
 */

#define VFD_READ	(0x01)
#define VFD_WRITE	(0x02)
#define VFD_TIMEOUT	(0x04)

typedef struct vfd_t vfd_t;
typedef void (*vfd_callback_t)(void *uap, int eventmask);
typedef struct { void *opaque; } vfd_evid_t;

typedef enum {
	/* get only -- void *value is a pointer to int */
	VFD_SOCKERR,
} vfd_sockopt_t;

/* XXX: replace errno with our own error mechanism, bleh */
typedef struct {
	/* read uses the buffers pointed to by buf_tmp to read as many
	 * bytes as possible -- it splits the bytes read off into buf_result
	 * and releases them from buf_tmp.
	 *
	 * amt_read must not be NULL, it will be filled in with the number of
	 * bytes read.
	 *
	 * the return is 1 for OK, 0 for EOF, and -1 for error with
	 * errno set.  it is possible for EOF or error to occur in addition
	 * to bytes being read.
	 */
	int (*read)(vfd_t *, zc_t *buf_tmp, zc_t *buf_result, size_t *amt_read);

	/* write writes as many bytes from buf as possible without
	 * blocking.  zc_consume_bytes is called to release the
	 * data which has been written.
	 *
	 * amt_written must not be NULL, it will be filled in with the number of
	 * bytes written.
	 *
	 * the return is 0 for no error, or -1 for error with errno set
	 * it's possible for an error to be returned in addition to
	 * bytes being written.
	 */
	int (*write)(vfd_t *, zc_t *buf, size_t *amt_written);

	/* select registers a callback which will be invoked when the
	 * vfd is ready for the indicated events.
	 *
	 * the callback is triggered at most once; and must be re-registered
	 * if further events are desired.
	 *
	 * it is possible to register multiple callbacks for the same vfd.
	 *
	 * the events are a bitmask of VFD_READ, VFD_WRITE,
	 * and VFD_TIMEOUT.
	 *
	 * timeout indicates the timeout in microseconds, VFD_TIMEOUT
	 * must also have been set in the bitmask.
	 *
	 * callback is called when one of the events occured.
	 *
	 * if id is non-NULL then it is set to a token which can be used by
	 * deselect to later unregister the event.
	 *
	 * return 0 on success -1/errno on error
	 */
	int (*select)(vfd_t *, int eventmask, int timeout,
		void *uap, vfd_callback_t callback, vfd_evid_t *id);

	/* deselect a set of events for a vfd
	 *
	 * return 0 on success -1/errno on error
	 */
	int (*deselect)(vfd_evid_t *id);

	/* shutdown communication in the indicated direction
	 * how = 0   shutdown receive
	 * how = 1   shutdown send
	 * how = 2   shutdown send and receive
	 */
	int (*shutdown)(vfd_t *, int how);

	/* close closes the underlying transport, and frees the vfd structure.
	 * the caller must ensure that there are no outstanding registered events.
	 */
	int (*close)(vfd_t *);

	/* accept accepts a new connection
	 * XXX: need to handle the peer addr and such
	 */
	vfd_t *(*accept)(vfd_t *);

	/* getsockopt */
	int (*getsockopt)(vfd_t *, vfd_sockopt_t, void *);

	/* setsockopt */
	int (*setsockopt)(vfd_t *, vfd_sockopt_t, void *);
} vfd_methods_t;

/* this structure should be considered private */
struct vfd_t {
	const vfd_methods_t *methods;
	splim_t *ctx;
};

#define vfd_read(_v, _t, _b, _a) ((_v)->methods->read((_v), (_t), (_b), (_a)))
#define vfd_write(_v, _b, _a) ((_v)->methods->write((_v), (_b), (_a)))
#define vfd_select(_v, _e, _t, _f, _u, _i) ((_v)->methods->select((_v), (_e), (_t), (_f), (_u), (_i)))
#define vfd_deselect(_v, _i) ((_v)->methods->deselect((_v), (_i)))
#define vfd_shutdown(_v, _h) ((_v)->methods->shutdown((_v), (_h)))
#define vfd_close(_v) ((_v)->methods->close((_v)))
#define vfd_accept(_v) ((_v)->methods->accept((_v)))
#define vfd_getsockopt(_v, _o, _s) ((_v)->methods->getsockopt((_v), (_o), (_s)))
#define vfd_setsockopt(_v, _o, _s) ((_v)->methods->setsockopt((_v), (_o), (_s)))

/* create a vfd for the indicated socket */
vfd_t *vfd_socket(splim_t *ctx, int fd);

#endif
