/* ====================================================================
 * Copyright (c) 2000 Dean Gaudet.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Dean Gaudet"
 *
 * THIS SOFTWARE IS PROVIDED BY DEAN GAUDET ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DEAN GAUDET OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */
#ifndef SPLIM_ZC_H
#define SPLIM_ZC_H 1

/*
	zc is a zero-copy buffering abstraction.

	it's based on an abstraction which Jason Evans
	<jasone@canonware.com> wrote in a library called libstash.
	Jason's code was far too heavy-weight for my tastes.  what i have
	here is very minimal -- and i am very deliberately exposing the
	structures to the callers.

	i'm not sure that this is the only or best abstraction for
	doing zero-copy.  i know there are others i could study (such
	as skbuffs in the linux kernel), but for now i'm just going
	to use this.
*/

#include <stddef.h>
#include <sys/uio.h>

/*
	the actual data goes into typical char *buf; arrays, but in
	order to use such a buf you need to set up a bufc_t to contain
	the accounting information for that buffer.  specifically the
	reference count; and a function to free the data when the ref
	count drops to zero.

	note that if ref_count > 1 then the buffer is shared and
	shouldn't be modified.
	TODO: API function to prepare for write by copying if necessary
*/
typedef struct bufc_t bufc_t;
struct bufc_t {
	char *buf;
	unsigned ref_count;
	void (*free)(void *uap, void *buf);
	void *uap;
};

/*
	bufel_t are lists of bufc_ts, plus an offset and length.
	in this way you can refer to all or a subset of a bufc_t.
*/
typedef struct bufel_t bufel_t;
struct bufel_t {
	bufel_t *bufel_next;
	bufel_t *bufel_prev;
	bufc_t *bufc;
	char *start;	/* points within the bufc to the first byte */
	size_t len;	/* number of bytes in element */
};

ring_template(bufel_t, bufel)

/*
	a zc_t is essentially a list of bufel_ts.
	
	the ctx is used for memory allocation when required.
	the functions taking multiple zc_ts below assume that
	all the zc_ts have the same ctx.  this goes back to the
	single-threaded philosophy.
*/
typedef struct {
	bufel_ring_t elts;
	splim_t *ctx;
} zc_t;

/* initialise a zc for use */
extern void zc_init(splim_t *, zc_t *);

/* finalise a zc -- free up any buffers and such */
extern void zc_fini(zc_t *);

/* re-initialise a zc -- like calling zc_fini then zc_init */
extern void zc_reinit(zc_t *);

/* the number of bytes referenced by the zc_t */
extern size_t zc_length(zc_t *);

/* true if the zc_t has no bytes */
extern int zc_is_empty(zc_t *);

/* remove bytes from the head of the zc */
extern void zc_consume_bytes(zc_t *, size_t bytes);

/* turn the buffer list src into an array of iovecs dest
 * nvec is the maximum number of elements to return
 * result code is the number of elements used
 */
extern size_t zc_make_iovec(struct iovec *dest, size_t nvec, zc_t *src);

/* bufc_new counts for one reference */
extern bufc_t *bufc_new(splim_t *, char *buf, void *uap, void (*free)(void *uap, void *buf));

/* bufc_deref dereferences bufc by 1, and frees if it's the last ref */
extern void bufc_deref(splim_t *, bufc_t *);

/* XXX: don't think this needs to be public */
extern bufel_t *bufel_new(splim_t *, bufc_t *bufc, size_t start, size_t len);

/* create the bufc and bufel for buf, and append it to zc */
extern void zc_append_buf(zc_t *zc, void *buf, size_t start, size_t len, void *uap, void (*free)(void *uap, void *buf));

/* append a bufc to zc */
extern void zc_append_bufc(zc_t *zc, bufc_t *, size_t start, size_t len);

/* input:
 * A = x..p q..y        where q is the (offset)th byte
 * B = u..v
 *
 * output:
 * A = q..y
 * B = u..v x..p
 */
extern void zc_split(zc_t *B, zc_t *A, size_t offset);

/* until the rest of the world catches up with us and provides
 * APIs that take zc_ts, we need this convenience function to
 * copy a zc_t into a linear buffer.  (of course defeating the
 * whole point of zero copy :)
 */
extern void zc_linear(void *dest, zc_t *src, size_t len);

#endif
