/* ====================================================================
 * Copyright (c) 2000 Dean Gaudet.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Dean Gaudet"
 *
 * THIS SOFTWARE IS PROVIDED BY DEAN GAUDET ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DEAN GAUDET OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */
#include <splim/splim.h>
#include <stdint.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>

#include "splim/splim_p.h"


void kev_ctx_init(splim_t *ctx)
{
	io_ring_init(&ctx->kev.ios);
	timeout_ring_init(&ctx->kev.timeouts);
	ctx->kev.io_pezz = pezz_create(sizeof(kev_io_t), HAPPY_PEZZ);
	ctx->kev.timeout_pezz = pezz_create(sizeof(kev_timeout_t), HAPPY_PEZZ);
}

void kev_ctx_fini(splim_t *ctx)
{
	/* all memory is freed by destroying the pezzs */
	pezz_destroy(ctx->kev.io_pezz);
	pezz_destroy(ctx->kev.timeout_pezz);
}


kev_io_t *kev_register_io(splim_t *ctx, int fd, int events, void *uap, void (*callback)(void *, int))
{
	kev_io_t *io;

	io = pezz_alloc(ctx->kev.io_pezz);
	io->fd = fd;
	io->events = events;
	io->uap = uap;
	io->callback = callback;
	io_append(&ctx->kev.ios, io);
	return io;
}


void kev_unregister_io(kev_io_t *io)
{
	/* delayed cleanup to avoid generic nastiness in kev_main_loop
	 * if a callback unregisters other ios
	 */
	io->callback = NULL;
}


kev_timeout_t *kev_register_timeout(splim_t *ctx, int microseconds, void *uap, void (*callback)(void *uap))
{
	kev_timeout_t *timeout;
	kev_timeout_t *timeout_cur;
	kev_timeout_t *timeout_next;
	kev_timeout_t *timeout_stop;

	timeout = pezz_alloc(ctx->kev.timeout_pezz);
	timeout->uap = uap;
	timeout->callback = callback;
	timeout->trigger_time = ctx->kev.last_event + microseconds;

	for (timeout_cur = timeout_first(&ctx->kev.timeouts), timeout_stop = timeout_sentinel(&ctx->kev.timeouts)
		; timeout_cur != timeout_stop
		; timeout_cur = timeout_next) {

		timeout_next = timeout_cur->timeout_next;

		/* XXX: hrm, <= or < here?  <= causes LIFO action, but avoids O(n^2) wonder what that might do? -dean */
		if (timeout->trigger_time <= timeout_cur->trigger_time) {
			timeout_insert_before(timeout, timeout_cur);
			return timeout;
		}
	}
	/* ring must be empty */
	timeout_append(&ctx->kev.timeouts, timeout);
	return timeout;
}


void kev_unregister_timeout(kev_timeout_t *timeout)
{
	/* delayed cleanup to avoid generic nastiness in kev_main_loop
	 * if a callback unregisters other timeouts
	 */
	timeout->callback = NULL;
}


static kev_time_t kev_now(void)
{
	struct timeval tv;
	gettimeofday(&tv, 0);
	return (kev_time_t)tv.tv_sec * INT64_C(1000000) + (kev_time_t)tv.tv_usec;
}


int kev_main_loop(splim_t *ctx)
{
	kev_time_t next_timeout;
	kev_timeout_t *timeout_cur;
	kev_timeout_t *timeout_next;
	kev_timeout_t *timeout_stop;
	kev_io_t *io_cur;
	kev_io_t *io_next;
	kev_io_t *io_stop;
	struct pollfd *fd_set;
	unsigned n_fds;
	int fd_index;
	int timeout;
	int rc;
	int i;

	n_fds = 1;
	fd_set = malloc(sizeof(struct pollfd));
	for (;;) {
		/* check the timeouts */
		ctx->kev.last_event = kev_now();
		next_timeout = 0;
		for (timeout_cur = timeout_first(&ctx->kev.timeouts), timeout_stop = timeout_sentinel(&ctx->kev.timeouts)
			; timeout_cur != timeout_stop
			; timeout_cur = timeout_next) {

			timeout_next = timeout_cur->timeout_next;

			if (timeout_cur->callback == NULL) {
				timeout_unlink(timeout_cur);
				pezz_free(ctx->kev.timeout_pezz, timeout_cur);
			}
			else if (timeout_cur->trigger_time > ctx->kev.last_event) {
				next_timeout = timeout_cur->trigger_time;
				break;
			}
			else {
				timeout_unlink(timeout_cur);
				timeout_cur->callback(timeout_cur->uap);
				pezz_free(ctx->kev.timeout_pezz, timeout_cur);
			}
		}

		/* generate the fd_set array */
		fd_index = 0;
		for (io_cur = io_first(&ctx->kev.ios), io_stop = io_sentinel(&ctx->kev.ios)
			; io_cur != io_stop
			; io_cur = io_next) {

			io_next = io_cur->io_next;

			if (io_cur->callback == NULL) {
				io_unlink(io_cur);
				pezz_free(ctx->kev.io_pezz, io_cur);
			}
			else {
				if (fd_index >= n_fds) {
					n_fds = n_fds * 2;
					fd_set = realloc(fd_set, sizeof(struct pollfd) * n_fds);
				}
				fd_set[fd_index].fd = io_cur->fd;
				fd_set[fd_index].events = io_cur->events;
				++fd_index;
			}
		}

		timeout = -1;
		if (next_timeout) {
			timeout = (next_timeout - ctx->kev.last_event) / 1000;
		}

		rc = poll(fd_set, fd_index, timeout);

		/* we iterate over both the ios ring and the fd_set in
		 * lock-step.  the callbacks are free to unregister ios
		 * or register new ios -- neither case violates the
		 * careful arrangement we made to keep the fd_set and
		 * the ring in sync.  this is because we add new ios to
		 * the end of the ring; and when ios are unregistered we
		 * delay unlinking them until we're constructing the
		 * next fd_set
		 */
		i = 0;
		io_cur = io_first(&ctx->kev.ios);
		while (rc > 0) {
			io_next = io_cur->io_next;
			if (fd_set[i].revents) {
				--rc;
				if (io_cur->callback) {
					io_unlink(io_cur);
					io_cur->callback(io_cur->uap, fd_set[i].revents);
					pezz_free(ctx->kev.io_pezz, io_cur);
				}
			}
			io_cur = io_next;
			++i;
		}

		if (rc < 0) {
			if (errno != EINTR) {
				/* uh oh */
				int save_errno = errno;
				free(fd_set);
				errno = save_errno;
				return -1;
			}
		}
	}
}
