/* ====================================================================
 * Copyright (c) 2000 Dean Gaudet.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Dean Gaudet"
 *
 * THIS SOFTWARE IS PROVIDED BY DEAN GAUDET ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DEAN GAUDET OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */
#include <splim/splim.h>
#include <stdlib.h>

#include "splim/splim_p.h"

/* if set to 1 then we disable the pezz stuff so that an
 * external memory debugging tool can work.
 */
#define PEZZ_DISABLE 0

typedef struct dispenser_t dispenser_t;
struct dispenser_t {
	dispenser_t *next;
	char data[0];	/* XXX: is alignment required? */
};

struct pezz_t {
	size_t element_size;
	size_t dispenser_size;
	void **free_list;
	dispenser_t *dispensers;
	char *top_elm;
	size_t num_allocated;
	size_t num_in_use;
};


extern pezz_t *pezz_create(size_t element_size, size_t dispenser_size)
{
	pezz_t *pezz;

	pezz = malloc(sizeof(*pezz));
	pezz->element_size = element_size;
	pezz->dispenser_size = dispenser_size;
	pezz->free_list = NULL;
	pezz->dispensers = NULL;
	pezz->top_elm = NULL;
	pezz->num_allocated = 0;
	pezz->num_in_use = 0;
	return pezz;
}


void pezz_destroy(pezz_t *pezz)
{
	dispenser_t *next;
	dispenser_t *cur;

	for (cur = pezz->dispensers; cur; cur = next) {
		next = cur->next;
		free(cur);
	}
	free(pezz);
}


void *pezz_alloc(pezz_t *pezz)
{
#if PEZZ_DISABLE
	return malloc(pezz->element_size);
#else
	void *result;
	dispenser_t *d;
	char *cur;
	char *end;
	char *next;

	if (!pezz->free_list) {
		/* create a new dispenser */
		d = malloc(sizeof(dispenser_t) + pezz->element_size * pezz->dispenser_size);
		d->next = pezz->dispensers;
		pezz->dispensers = d;

		/* and link all of its elements onto the free_list */
		cur = &d->data[0];
		end = cur + pezz->element_size * (pezz->dispenser_size - 1);
		while (cur < end) {
			next = cur + pezz->element_size;
			*(void **)cur = next;
			cur = next;
		}

		*(void **)cur = NULL;
		pezz->free_list = (void **)&d->data[0];

		pezz->num_allocated += pezz->dispenser_size;
	}
	++pezz->num_in_use;
	result = pezz->free_list;
	pezz->free_list = *(void **)result;
	return result;
#endif
}


void pezz_free(pezz_t *pezz, void *v)
{
#if PEZZ_DISABLE
	free(v);
#else
	*(void **)v = pezz->free_list;
	pezz->free_list = (void **)v;
	--pezz->num_in_use;
#endif
}


void pezz_bufc_free(void *uap, void *buf)
{
	pezz_t *pezz = (pezz_t *)uap;

	pezz_free(pezz, buf);
}
