/* ====================================================================
 * Copyright (c) 2000 Dean Gaudet.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Dean Gaudet"
 *
 * THIS SOFTWARE IS PROVIDED BY DEAN GAUDET ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DEAN GAUDET OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */
#include <splim/splim.h>
#include <errno.h>
#include <unistd.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <netinet/tcp.h>
#include <netinet/in.h>

#include "splim/splim_p.h"


typedef struct {
	vfd_t vfd;
	int fd;
} vfd_socket_t;

static int vfd_socket_read(vfd_t *vvfd, zc_t *buf_tmp, zc_t *buf_result, size_t *amt_read)
{
	vfd_socket_t *vfd = (vfd_socket_t *)vvfd;
	struct iovec vec[UIO_MAXIOV];
	int nvec;
	int bytes_read;

	nvec = zc_make_iovec(vec, UIO_MAXIOV, buf_tmp);
	do {
		bytes_read = readv(vfd->fd, vec, nvec);
	} while (bytes_read < 0 && errno == EINTR);
	if (bytes_read > 0) {
		*amt_read = bytes_read;
		zc_split(buf_result, buf_tmp, bytes_read);
		return 1;
	}
	*amt_read = 0;
	return bytes_read;
}


static int vfd_socket_write(vfd_t *vvfd, zc_t *buf, size_t *amt_written)
{
	vfd_socket_t *vfd = (vfd_socket_t *)vvfd;
	struct iovec vec[UIO_MAXIOV];
	int nvec;
	int bytes_written;

	nvec = zc_make_iovec(vec, UIO_MAXIOV, buf);
	do {
		bytes_written = writev(vfd->fd, vec, nvec);
	} while (bytes_written < 0 && errno == EINTR);
	if (bytes_written > 0) {
		*amt_written = bytes_written;
		zc_consume_bytes(buf, bytes_written);
		return 0;
	}
	*amt_written = 0;
	return bytes_written;
}


typedef struct {
	vfd_socket_t *vfd;
	vfd_callback_t callback;
	void *uap;
	kev_timeout_t *timer_id;
	kev_io_t *io_id;
} vfd_socket_evid_t;


static void vfd_socket_timeout(void *uap)
{
	vfd_socket_evid_t *id = uap;

	kev_unregister_io(id->io_id);
	id->callback(id->uap, VFD_TIMEOUT);
	pezz_free(id->vfd->vfd.ctx->vfd_socket.evid_pezz, id);
}


static void vfd_socket_io(void *uap, int poll_events)
{
	vfd_socket_evid_t *id = uap;
	int events;

	if (id->timer_id) {
		kev_unregister_timeout(id->timer_id);
	}
	events = 0;
	if (poll_events & (POLLIN|POLLHUP|POLLERR)) events |= VFD_READ;
	if (poll_events & (POLLOUT|POLLHUP|POLLERR)) events |= VFD_WRITE;
	id->callback(id->uap, events);
	pezz_free(id->vfd->vfd.ctx->vfd_socket.evid_pezz, id);
}


static int vfd_socket_select(vfd_t *vvfd, int eventmask, int timeout,
	void *uap, vfd_callback_t callback, vfd_evid_t *pid)
{
	vfd_socket_t *vfd = (vfd_socket_t *)vvfd;
	vfd_socket_evid_t *id;
	int poll_events;

	id = pezz_alloc(vfd->vfd.ctx->vfd_socket.evid_pezz);
	if (pid) {
		pid->opaque = id;
	}

	id->vfd = vfd;
	id->callback = callback;
	id->uap = uap;
	id->timer_id = NULL;

	poll_events = POLLHUP | POLLERR;
	if (eventmask & VFD_READ) poll_events |= POLLIN;
	if (eventmask & VFD_WRITE) poll_events |= POLLOUT;

	id->io_id = kev_register_io(vfd->vfd.ctx, vfd->fd, poll_events, id, vfd_socket_io);
	if (id->io_id == NULL) {
		int save_errno = errno;
		pezz_free(vfd->vfd.ctx->vfd_socket.evid_pezz, id);
		errno = save_errno;
		return -1;
	}

	if (eventmask & VFD_TIMEOUT) {
		id->timer_id = kev_register_timeout(vfd->vfd.ctx, timeout, id, vfd_socket_timeout);
		if (id->timer_id == NULL) {
			int save_errno = errno;
			kev_unregister_io(id->io_id);
			pezz_free(vfd->vfd.ctx->vfd_socket.evid_pezz, id);
			errno = save_errno;
			return -1;
		}
	}
	return 0;
}


static int vfd_socket_deselect(vfd_evid_t *vid)
{
	vfd_socket_evid_t *id = (vfd_socket_evid_t *)vid;

	kev_unregister_io(id->io_id);
	if (id->timer_id) {
		kev_unregister_timeout(id->timer_id);
	}
	pezz_free(id->vfd->vfd.ctx->vfd_socket.evid_pezz, id);
	return 0;
}


static int vfd_socket_shutdown(vfd_t *vvfd, int how)
{
	vfd_socket_t *vfd = (vfd_socket_t *)vvfd;

	return shutdown(vfd->fd, how);
}


static int vfd_socket_close(vfd_t *vvfd)
{
	vfd_socket_t *vfd = (vfd_socket_t *)vvfd;
	int rc;
	int save_errno;

	rc = close(vfd->fd);
	save_errno = errno;
	pezz_free(vfd->vfd.ctx->vfd_socket.vfd_socket_pezz, vfd);
	errno = save_errno;
	return rc;
}


static vfd_t *vfd_socket_accept(vfd_t *vvfd)
{
	vfd_socket_t *vfd = (vfd_socket_t *)vvfd;
	int new_fd;
	int just_say_no;

	do {
		new_fd = accept(vfd->fd, NULL, NULL);
	} while (new_fd < 0 && errno == EINTR);
	if (new_fd < 0) {
		return NULL;
	}

	just_say_no = 1;
	if (setsockopt(new_fd, IPPROTO_TCP, TCP_NODELAY,
		(char *)&just_say_no, sizeof(just_say_no)) < 0) {
		close(new_fd);
		return NULL;
	}

	return vfd_socket(vfd->vfd.ctx, new_fd);
}


static int vfd_socket_setsockopt(vfd_t *vvfd, vfd_sockopt_t opt, void *value)
{
	errno = EINVAL;
	return -1;
}


static int vfd_socket_getsockopt(vfd_t *vvfd, vfd_sockopt_t opt, void *value)
{
	vfd_socket_t *vfd = (vfd_socket_t *)vvfd;
	int sockerr_len;

	switch (opt) {
	case VFD_SOCKERR:
		sockerr_len = sizeof(int);
		if (getsockopt(vfd->fd, SOL_SOCKET, SO_ERROR, value, &sockerr_len)) {
			return -1;
		}
		return 0;
	default:
		errno = EINVAL;
		return -1;
	}
}


const vfd_methods_t vfd_socket_methods = {
	.read = vfd_socket_read,
	.write = vfd_socket_write,
	.select = vfd_socket_select,
	.deselect = vfd_socket_deselect,
	.shutdown = vfd_socket_shutdown,
	.close = vfd_socket_close,
	.accept = vfd_socket_accept,
	.setsockopt = vfd_socket_setsockopt,
	.getsockopt = vfd_socket_getsockopt,
};


vfd_t *vfd_socket(splim_t *ctx, int fd)
{
	vfd_socket_t *vfd;

	if (fcntl(fd, F_SETFL, O_NONBLOCK | O_RDWR) < 0) {
		return NULL;
	}

	vfd = pezz_alloc(ctx->vfd_socket.vfd_socket_pezz);
	vfd->vfd.methods = &vfd_socket_methods;
	vfd->vfd.ctx = ctx;
	vfd->fd = fd;

	return (vfd_t *)vfd;
}


void vfd_socket_init(splim_t *ctx)
{
	ctx->vfd_socket.evid_pezz = pezz_create(sizeof(vfd_socket_evid_t), HAPPY_PEZZ);
	ctx->vfd_socket.vfd_socket_pezz = pezz_create(sizeof(vfd_socket_t), HAPPY_PEZZ);
}


void vfd_socket_fini(splim_t *ctx)
{
	pezz_destroy(ctx->vfd_socket.evid_pezz);
	pezz_destroy(ctx->vfd_socket.vfd_socket_pezz);
}
