/* ====================================================================
 * Copyright (c) 2000 Dean Gaudet.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Dean Gaudet"
 *
 * THIS SOFTWARE IS PROVIDED BY DEAN GAUDET ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DEAN GAUDET OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */
#include <splim/splim.h>
#include <assert.h>

#include "splim/splim_p.h"


void zc_ctx_init(splim_t *ctx)
{
	ctx->zc.bufc_pezz = pezz_create(sizeof(bufc_t), HAPPY_PEZZ);
	ctx->zc.bufel_pezz = pezz_create(sizeof(bufel_t), HAPPY_PEZZ);
}


void zc_ctx_fini(splim_t *ctx)
{
	pezz_destroy(ctx->zc.bufc_pezz);
	pezz_destroy(ctx->zc.bufel_pezz);
}


void zc_init(splim_t *ctx, zc_t *zc)
{
	bufel_ring_init(&zc->elts);
	zc->ctx = ctx;
}


size_t zc_length(zc_t *zc)
{
	bufel_t *cur;
	bufel_t *stop;
	size_t len;

	len = 0;
	for (cur = bufel_first(&zc->elts), stop = bufel_sentinel(&zc->elts)
		; cur != stop
		; cur = cur->bufel_next) {
		len += cur->len;
	}
	return len;
}


int zc_is_empty(zc_t *zc)
{
	return bufel_is_empty(&zc->elts);
}


void bufc_deref(splim_t *ctx, bufc_t *bufc)
{
	if (--bufc->ref_count == 0) {
		bufc->free(bufc->uap, bufc->buf);
		pezz_free(ctx->zc.bufc_pezz, bufc);
	}
}


static void bufel_free(zc_t *zc, bufel_t *bufel)
{
	bufc_deref(zc->ctx, bufel->bufc);
	bufel_unlink(bufel);
	pezz_free(zc->ctx->zc.bufel_pezz, bufel);
}


void zc_fini(zc_t *zc)
{
	bufel_t *cur;
	bufel_t *next;
	bufel_t *stop;

	for (cur = bufel_first(&zc->elts), stop = bufel_sentinel(&zc->elts)
		; cur != stop
		; cur = next) {
		next = cur->bufel_next;
		bufel_free(zc, cur);
	}
}


void zc_reinit(zc_t *zc)
{
	zc_fini(zc);
}


void zc_consume_bytes(zc_t *zc, size_t bytes)
{
	bufel_t *cur;
	bufel_t *next;
	bufel_t *stop;

	for (cur = bufel_first(&zc->elts), stop = bufel_sentinel(&zc->elts)
		; bytes > 0 && cur != stop
		; cur = next) {
		next = cur->bufel_next;
		if (bytes >= cur->len) {
			bytes -= cur->len;
			bufel_free(zc, cur);
		}
		else {
			cur->start += bytes;
			cur->len -= bytes;
			break;
		}
	}
}


size_t zc_make_iovec(struct iovec *dest, size_t nvec, zc_t *src)
{
	bufel_t *cur;
	bufel_t *stop;
	int i;

	i = 0;
	cur = bufel_first(&src->elts);
	stop = bufel_sentinel(&src->elts);
	while (i < nvec && cur != stop) {
		dest[i].iov_base = (void *)cur->start;
		dest[i].iov_len = cur->len;
		cur = cur->bufel_next;
		++i;
	}
	return i;
}


bufc_t *bufc_new(splim_t *ctx, char *buf, void *uap, void (*free)(void *, void *))
{
	bufc_t *bufc;

	bufc = pezz_alloc(ctx->zc.bufc_pezz);
	bufc->buf = buf;
	bufc->uap = uap;
	bufc->free = free;
	bufc->ref_count = 1;
	return bufc;
}


bufel_t *bufel_new(splim_t *ctx, bufc_t *bufc, size_t start, size_t len)
{
	bufel_t *bufel;

	bufel = pezz_alloc(ctx->zc.bufel_pezz);
	bufel_elt_init(bufel);
	bufel->bufc = bufc;
	++bufc->ref_count;
	bufel->start = bufc->buf + start;
	bufel->len = len;
	return bufel;
}


void zc_append_buf(zc_t *zc, void *buf, size_t start, size_t len, void *uap, void (*free)(void *, void *))
{
	bufc_t *bufc;
	bufel_t *bufel;

	if (len == 0) {
		return;
	}
	bufc = bufc_new(zc->ctx, buf, uap, free);
	--bufc->ref_count;
	bufel = bufel_new(zc->ctx, bufc, start, len);
	bufel_append(&zc->elts, bufel);
}


void zc_append_bufc(zc_t *zc, bufc_t *bufc, size_t start, size_t len)
{
	bufel_t *bufel;

	if (len == 0) {
		return;
	}
	bufel = bufel_new(zc->ctx, bufc, start, len);
	bufel_append(&zc->elts, bufel);
}


/* input:
 * A = x..p q..y        where q is the (offset)th byte
 * B = u..v
 *
 * output:
 * A = q..y
 * B = u..v x..p
 */
void zc_split(zc_t *B, zc_t *A, size_t offset)
{
	bufel_t *cur;
	bufel_t *stop;
	bufel_t *new;

	assert(A->ctx == B->ctx);

	if (offset == 0) {
		return;
	}

	/* find the bufel to split at */
	for (cur = bufel_first(&A->elts), stop = bufel_sentinel(&A->elts)
		; cur != stop && offset > cur->len
		; cur = cur->bufel_next) {
		offset -= cur->len;
	}

	/* offset was beyond the end of A ... just split everything */
	if (cur == stop) {
		if (!bufel_is_empty(&A->elts)) {
			bufel_splice_after(bufel_last(&B->elts), bufel_first(&A->elts), bufel_last(&A->elts));
		}
		return;
	}

	/* move the bufels to B */
	bufel_splice_after(bufel_last(&B->elts), bufel_first(&A->elts), cur);

	/* offset conveniently fits on a bufel boundary */
	if (offset == cur->len) {
		return;
	}

	/* otherwise need to create new bufel and adjust stuff */
	new = bufel_new(A->ctx, cur->bufc, (cur->start - cur->bufc->buf) + offset, cur->len - offset);
	cur->len = offset;
	bufel_prepend(&A->elts, new);
}


void zc_linear(void *vdest, zc_t *src, size_t len)
{
	bufel_t *cur;
	bufel_t *stop;
	char *dest = (char *)vdest;

	for (cur = bufel_first(&src->elts), stop = bufel_sentinel(&src->elts)
		; cur != stop
		; cur = cur->bufel_next) {
		if (len > cur->len) {
			memcpy(dest, cur->start, cur->len);
			dest += cur->len;
			len -= cur->len;
		}
		else {
			memcpy(dest, cur->start, len);
			return;
		}
	}
}
