/* ====================================================================
 * Copyright (c) 2000 Dean Gaudet.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Dean Gaudet"
 *
 * THIS SOFTWARE IS PROVIDED BY DEAN GAUDET ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DEAN GAUDET OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */
#include <splim/splim.h>
#include <stdlib.h>
#include <netdb.h>
#include <stdio.h>
#include <signal.h>
#include <unistd.h>

unsigned num_clients;

struct sockaddr_in server_addr;
splim_t *ctx;

#define BUFFER_SIZE (4096)
unsigned char two_five_six[BUFFER_SIZE];

typedef struct client_t client_t;
struct client_t {
	vfd_t *vfd;
	unsigned expected_amount;
	unsigned amount_received;
};

pezz_t *client_pezz;

uint32_t stats_connect_errors;
uint32_t stats_write_errors;
uint32_t stats_wrong_data;
uint32_t stats_read_error;
uint32_t stats_finished_properly;
uint32_t stats_improper_length;
uint64_t stats_bytes_received;

pezz_t *request_pezz;
zc_t request_zc;
zc_t page_buffer;
zc_t read_buffer;

/* forward declaration, state machines are fun! */
static void launch_client(void);


static void delete_client(client_t *c)
{
	vfd_close(c->vfd);
	pezz_free(client_pezz, c);

	/* recreate the client immediately */
	launch_client();
}


static void read_response(void *vc, int eventmask)
{
	client_t *c = vc;
	int rc;
	size_t amt_read;
	bufel_t *cur;
	bufel_t *stop;

	pagemill(&page_buffer);

	rc = vfd_read(c->vfd, &page_buffer, &read_buffer, &amt_read);
	if (rc < 0) {
		++stats_read_error;
		delete_client(c);
		return;
	}
	stats_bytes_received += amt_read;

#if 1
	if (amt_read % 32) {
		zc_consume_bytes(&page_buffer, 32 - (amt_read % 32));
	}
#endif

	for (cur = bufel_first(&read_buffer.elts), stop = bufel_sentinel(&read_buffer.elts)
		; cur != stop
		; cur = cur->bufel_next) {

		unsigned amt;
		int i;
		int sludge;

		sludge = c->amount_received & 0xff;
		amt = cur->len;
		c->amount_received += amt;
		i = 0;
		while (amt >= 256) {
			if (memcmp(cur->start + i, two_five_six + sludge, 256)) {
doom:
				++stats_wrong_data;
				delete_client(c);
				zc_reinit(&read_buffer);
				return;
			}
			i += 256;
			amt -= 256;
		}

		if (memcmp(cur->start + i, two_five_six + sludge, amt)) {
			goto doom;
		}
	}

	zc_reinit(&read_buffer);

	/* handle eof */
	if (rc == 0) {
		if (c->amount_received == c->expected_amount) {
			++stats_finished_properly;
		}
		else {
			++stats_improper_length;
		}
		delete_client(c);
		return;
	}

	vfd_select(c->vfd, VFD_READ, 0, c, read_response, NULL);
}


static void connected(void *vc, int eventmask)
{
	client_t *c = vc;
	int sockerr;
	char *buf;
	size_t amt_written;
	int rc;

	if (vfd_getsockopt(c->vfd, VFD_SOCKERR, &sockerr)) {
		delete_client(c);
		return;
	}
	if (sockerr) {
		++stats_connect_errors;
		delete_client(c);
		return;
	}

	/* time to write the request */

	c->expected_amount = drand48() * 50000 + 1000;
	c->amount_received = 0;

	buf = pezz_alloc(request_pezz);
	/* the address of c is our test "mailbox name" for affinity testing */
	sprintf(buf, "%u XYZ%pXYZ", c->expected_amount, c);

	zc_append_buf(&request_zc, buf, 0, 256, request_pezz, pezz_bufc_free);

	rc = vfd_write(c->vfd, &request_zc, &amt_written);

	/* regardless of the result we want to zero-out request_zc */
	zc_reinit(&request_zc);

	if (rc) {
		++stats_write_errors;
		delete_client(c);
		return;
	}

	vfd_select(c->vfd, VFD_READ, 0, c, read_response, NULL);
}


static void launch_client(void)
{
	client_t *c;

	c = pezz_alloc(client_pezz);
	c->vfd = TCP.connect(ctx, &server_addr);

	vfd_select(c->vfd, VFD_WRITE, 0, c, connected, NULL);
}


static void handle_sig(int signum)
{
	printf("%40s: %15qu\n", "bytes_received", stats_bytes_received);
#define pstat(n)	\
	if (stats_##n) { \
		printf("%40s: %15u\n", #n, stats_##n); \
	}
	pstat(connect_errors);
	pstat(write_errors);
	pstat(wrong_data);
	pstat(read_error);
	pstat(finished_properly);
	pstat(improper_length);
	exit(0);
}


int main(int argc, char **argv)
{
	struct sigaction sa;
	struct hostent *hep;
	unsigned u;
	int c;
	char *argv0;
	int i;

	argv0 = argv[0];

	srand48(0);	/* consistency is best */

	for (u = 0; u < sizeof(two_five_six); ++u) {
		two_five_six[u] = u;
	}

	while ((c = getopt(argc, argv, "x")) != -1) {
		switch (c) {
		case 'x':
			/* i'm sure i'll want opts later */
			break;
		default:
			goto usage;
		}
	}
	argc -= optind;
	argv += optind;

	if (argc != 4) {
usage:
		fprintf(stderr,
			"usage: %s host port num_clients duration\n"
			, argv0);
		exit(1);
	}


	hep = gethostbyname(argv[0]);
	if (!hep || hep->h_addrtype != AF_INET || !hep->h_addr_list[0]) {
		perror("gethostbyname");
		exit(1);
	}
	memset(&server_addr, 0, sizeof(server_addr));
	server_addr.sin_family = AF_INET;
	server_addr.sin_addr = *(struct in_addr *)hep->h_addr_list[0];
	server_addr.sin_port = htons(atoi(argv[1]));

	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	sa.sa_handler = handle_sig;
	if (sigaction(SIGTERM, &sa, NULL) < 0) {
		perror("sigaction(SIGTERM)");
		exit(1);
	}
	if (sigaction(SIGINT, &sa, NULL) < 0) {
		perror("sigaction(SIGINT)");
		exit(1);
	}
	if (sigaction(SIGALRM, &sa, NULL) < 0) {
		perror("sigaction(SIGINT)");
		exit(1);
	}
	signal(SIGPIPE, SIG_IGN);

	num_clients = strtoul(argv[2], NULL, 0);

	alarm(atoi(argv[3]));

	ctx = splim_create();
	if (!ctx) {
		perror("splim_create");
		exit(1);
	}

	client_pezz = pezz_create(sizeof(client_t), HAPPY_PEZZ);
	request_pezz = pezz_create(256, HAPPY_PEZZ);

	zc_init(ctx, &request_zc);
	zc_init(ctx, &page_buffer);
	zc_init(ctx, &read_buffer);

	for (i = 0; i < num_clients; ++i) {
		launch_client();
	}

	return kev_main_loop(ctx);
}
