/* ====================================================================
 * Copyright (c) 2000 Dean Gaudet.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Dean Gaudet"
 *
 * THIS SOFTWARE IS PROVIDED BY DEAN GAUDET ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DEAN GAUDET OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */
#include <stdio.h>
#include <signal.h>
#include <splim/splim.h>
#include <stdint.h>
#include <assert.h>
#include <stdlib.h>
#include <errno.h>

splim_t *ctx;
vfd_t *main_socket;
vfd_evid_t main_event;
zc_t page_buffer;
zc_t read_buffer;

uint64_t stats_connections;
uint64_t stats_no_request;
uint64_t stats_write_error;
uint64_t stats_bytes_written;
uint64_t stats_responses;

typedef struct client_t client_t;
struct client_t {
	vfd_t *vfd;
	unsigned bytes_written;
	unsigned amount_to_write;
	zc_t response;
};

pezz_t *client_pezz;

#define BUFFER_SIZE (4096)
char bogus_buf[BUFFER_SIZE];
bufc_t *bogus_bufc;

/* no-op for bufc free */
static void bogus_free(void *a, void *b)
{
}


static void close_client(client_t *c)
{
	zc_fini(&c->response);
	vfd_close(c->vfd);
	pezz_free(client_pezz, c);
}


void bp(void) {}


static void write_response(void *uap, int eventmask)
{
	client_t *c = (client_t *)uap;
	int rc;
	size_t amt_written;

	assert(eventmask & VFD_WRITE);

	rc = vfd_write(c->vfd, &c->response, &amt_written);
	stats_bytes_written += amt_written;
	if (rc != 0) {
		++stats_write_error;
		close_client(c);
		return;
	}

	if (!zc_is_empty(&c->response)) {
		vfd_select(c->vfd, VFD_WRITE, 0, c, write_response, NULL);
		return;
	}

	++stats_responses;

	close_client(c);
}


void ready_for_request(void *uap, int eventmask)
{
	client_t *c = (client_t *)uap;
	int rc;
	char buf[2*BUFFER_SIZE];
	size_t amt_read;
	size_t amount_to_write;

	assert(eventmask & VFD_READ);
	pagemill(&page_buffer);

	rc = vfd_read(c->vfd, &page_buffer, &read_buffer, &amt_read);
	if (rc <= 0) {
		zc_reinit(&read_buffer);	/* free up the bytes we read */
		++stats_no_request;
		close_client(c);
		return;
	}

	assert(amt_read < sizeof(buf));
	zc_linear(buf, &read_buffer, amt_read);
#if 1
	//printf("amt_read = %u\n", amt_read);
	zc_consume_bytes(&read_buffer, amt_read);
	if (!zc_is_empty(&read_buffer)) {
		bp();
	}
#else
	zc_reinit(&read_buffer);	/* free up the bytes we read */
#endif

	/* concatenate a bunch of bogus_bufs up to the size we're supposed
	 * to respond with
	 */
	amount_to_write = atoi(buf);
	while (amount_to_write > BUFFER_SIZE) {
		zc_append_bufc(&c->response, bogus_bufc, 0, BUFFER_SIZE);
		amount_to_write -= BUFFER_SIZE;
	}
	zc_append_bufc(&c->response, bogus_bufc, 0, amount_to_write);

	vfd_select(c->vfd, VFD_WRITE, 0, c, write_response, NULL);
}

/* callback for when a new client arrives */
void new_client(void *uap, int eventmask)
{
	vfd_t *new_vfd;
	client_t *c;

	while ((new_vfd = vfd_accept(main_socket))) {
		++stats_connections;
		c = pezz_alloc(client_pezz);
		c->vfd = new_vfd;
		zc_init(ctx, &c->response);
		vfd_select(new_vfd, VFD_READ, 0, c, ready_for_request, NULL);
	}
	
	/* EAGAIN is expected if there are no more connections to accept */
	if (errno != EAGAIN) {
		perror("vfd_accept");
	}

	vfd_select(main_socket, VFD_READ, 0, NULL, new_client, &main_event);
}

/* SIGTERM handler */
void handle_sig(int s)
{
	printf("%40s: %15qu\n", "connections", stats_connections);
	printf("%40s: %15qu\n", "no_request", stats_no_request);
	printf("%40s: %15qu\n", "write_error", stats_write_error);
	printf("%40s: %15qu\n", "bytes_written", stats_bytes_written);
	printf("%40s: %15qu\n", "responses", stats_responses);
	exit(0);
}

int main(int argc, char **argv)
{
	struct sockaddr_in saddr;
	struct sigaction sa;
	int i;

	if (argc != 2) {
		fprintf(stderr, "usage: %s port\n", argv[0]);
		exit(1);
	}

	for (i = 0; i < BUFFER_SIZE; ++i) {
		bogus_buf[i] = (unsigned)i & 0xff;
	}

	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	sa.sa_handler = handle_sig;
	if (sigaction(SIGTERM, &sa, NULL) < 0) {
		perror("sigaction(SIGTERM)");
		exit(1);
	}
	if (sigaction(SIGINT, &sa, NULL) < 0) {
		perror("sigaction(SIGINT)");
		exit(1);
	}
	signal(SIGPIPE, SIG_IGN);

	ctx = splim_create();
	if (!ctx) {
		perror("splim_create");
		exit(1);
	}

	memset(&saddr, 0, sizeof(saddr));
	saddr.sin_family = AF_INET;
	saddr.sin_addr.s_addr = INADDR_ANY;
	saddr.sin_port = htons(atoi(argv[1]));
	main_socket = TCP.bind(ctx, &saddr);

	client_pezz = pezz_create(sizeof(client_t), HAPPY_PEZZ);
	zc_init(ctx, &page_buffer);
	zc_init(ctx, &read_buffer);

	bogus_bufc = bufc_new(ctx, bogus_buf, NULL, bogus_free);

	vfd_select(main_socket, VFD_READ, 0, NULL, new_client, &main_event);

	return kev_main_loop(ctx);
}
